/*
 * Test directory monitoring operations.
 *
 * Copyright (C) 2020, 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <thread>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>



/* Number of files to open in the test. */

static const int FILES_TO_OPEN = 10;



/* Open files in the directory given by filename. */

static void open_files(const char *filename)
{
  char *buffer = (char *) calloc(strlen(filename) + 10, sizeof(char));

  if (buffer == NULL)
  {
    printf("Could not allocate buffer.\n");
    return;
  }

  for (int i = 0; i < FILES_TO_OPEN; i++)
  {
    sprintf(buffer, "%s/file%02d", filename, i);

    printf("Open file %d.\n", i);
    client_open(buffer, O_CREAT | O_RDWR);
    sleep(1);
  }
}

/* Monitor files in the given directory. */

static void monitor_files(file_t *directory)
{
  int expected;

  for (expected = FILES_TO_OPEN; expected; expected--)
  {
    /* Wait for notification of content. */

    long err = client_wait_file(directory, client_notifier_task());

    if (err)
    {
      printf("Error waiting for notifications: %s\n", l4sys_errtostr(err));
      return;
    }

    printf("Notified with conditions:%s\n", file_notifications(directory) & NOTIFY_FILE_OPENED ? " file opened" : "");
  }

  printf("Notified for all expected files: %s\n", expected ? "False" : "True");
}



int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    printf("Need a directory name.\n");
    return 1;
  }

  char *filename = argv[1];

  printf("Opening %s...\n", filename);

  file_t *directory = client_open(filename, O_DIRECTORY);

  if (!client_opened(directory))
  {
    client_close(directory);
    printf("Could not open directory.\n");
    return 1;
  }

  long err = client_subscribe(directory, NOTIFY_FILE_OPENED, client_notifier_task());

  if (err)
  {
    printf("Could not subscribe to events on directory.\n");
    return 1;
  }

  /* Schedule threads. */

  std::thread *activities[2];

  activities[0] = new std::thread(open_files, filename);
  activities[1] = new std::thread(monitor_files, directory);

  for (int i = 0; i < 2; i++)
    activities[i]->join();

  printf("Expected %d 'file opened' notifications.\n", FILES_TO_OPEN);

  client_close(directory);

  printf("End of test.\n");
  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
